#XfceNotifyWindow {
    background-color: #4d4d4d;
    border-radius: 20px;
    border: 1px solid #fff;
    padding: 20px;
}

#XfceNotifyWindow:hover {
    border-width: 2px;
    border-color: #ffffff;
}

#XfceNotifyWindow label,
#XfceNotifyWindow image {
    color: #ffffff;
}

#XfceNotifyWindow label#summary {
    font-weight: Bold;
}

#XfceNotifyWindow button {
    border-style: none;
    background-image: none;
    background-color: #202020;
    color: #ffffff;
}

#XfceNotifyWindow button:hover {
    border-style: none;
    background-image: none;
    background-color: #404040;
    color: #ffffff;
}
