/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __XFCE_NOTIFY_LOG_GBUS_H__
#define __XFCE_NOTIFY_LOG_GBUS_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.xfce.Notifyd.Log */

#define XFCE_TYPE_NOTIFY_LOG_GBUS (xfce_notify_log_gbus_get_type ())
#define XFCE_NOTIFY_LOG_GBUS(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XFCE_TYPE_NOTIFY_LOG_GBUS, XfceNotifyLogGBus))
#define XFCE_IS_NOTIFY_LOG_GBUS(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XFCE_TYPE_NOTIFY_LOG_GBUS))
#define XFCE_NOTIFY_LOG_GBUS_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), XFCE_TYPE_NOTIFY_LOG_GBUS, XfceNotifyLogGBusIface))

struct _XfceNotifyLogGBus;
typedef struct _XfceNotifyLogGBus XfceNotifyLogGBus;
typedef struct _XfceNotifyLogGBusIface XfceNotifyLogGBusIface;

struct _XfceNotifyLogGBusIface
{
  GTypeInterface parent_iface;


  gboolean (*handle_clear) (
    XfceNotifyLogGBus *object,
    GDBusMethodInvocation *invocation);

  gboolean (*handle_delete) (
    XfceNotifyLogGBus *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *arg_ids);

  gboolean (*handle_get) (
    XfceNotifyLogGBus *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_id);

  gboolean (*handle_get_app_id_counts) (
    XfceNotifyLogGBus *object,
    GDBusMethodInvocation *invocation);

  gboolean (*handle_has_unread) (
    XfceNotifyLogGBus *object,
    GDBusMethodInvocation *invocation);

  gboolean (*handle_list) (
    XfceNotifyLogGBus *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_start_after_id,
    guint arg_count,
    gboolean arg_only_unread);

  gboolean (*handle_mark_all_read) (
    XfceNotifyLogGBus *object,
    GDBusMethodInvocation *invocation);

  gboolean (*handle_mark_read) (
    XfceNotifyLogGBus *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *arg_ids);

  gboolean (*handle_truncate) (
    XfceNotifyLogGBus *object,
    GDBusMethodInvocation *invocation,
    guint arg_entries_to_keep);

  void (*cleared) (
    XfceNotifyLogGBus *object);

  void (*row_added) (
    XfceNotifyLogGBus *object,
    const gchar *arg_id);

  void (*row_changed) (
    XfceNotifyLogGBus *object,
    const gchar *arg_id);

  void (*row_deleted) (
    XfceNotifyLogGBus *object,
    const gchar *arg_id);

  void (*truncated) (
    XfceNotifyLogGBus *object,
    guint arg_n_entries_kept);

};

GType xfce_notify_log_gbus_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *xfce_notify_log_gbus_interface_info (void);
guint xfce_notify_log_gbus_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void xfce_notify_log_gbus_complete_get (
    XfceNotifyLogGBus *object,
    GDBusMethodInvocation *invocation,
    GVariant *entry);

void xfce_notify_log_gbus_complete_list (
    XfceNotifyLogGBus *object,
    GDBusMethodInvocation *invocation,
    GVariant *entries);

void xfce_notify_log_gbus_complete_has_unread (
    XfceNotifyLogGBus *object,
    GDBusMethodInvocation *invocation,
    gboolean has_unread);

void xfce_notify_log_gbus_complete_get_app_id_counts (
    XfceNotifyLogGBus *object,
    GDBusMethodInvocation *invocation,
    GVariant *app_id_counts);

void xfce_notify_log_gbus_complete_mark_read (
    XfceNotifyLogGBus *object,
    GDBusMethodInvocation *invocation);

void xfce_notify_log_gbus_complete_mark_all_read (
    XfceNotifyLogGBus *object,
    GDBusMethodInvocation *invocation);

void xfce_notify_log_gbus_complete_delete (
    XfceNotifyLogGBus *object,
    GDBusMethodInvocation *invocation);

void xfce_notify_log_gbus_complete_truncate (
    XfceNotifyLogGBus *object,
    GDBusMethodInvocation *invocation);

void xfce_notify_log_gbus_complete_clear (
    XfceNotifyLogGBus *object,
    GDBusMethodInvocation *invocation);



/* D-Bus signal emissions functions: */
void xfce_notify_log_gbus_emit_row_added (
    XfceNotifyLogGBus *object,
    const gchar *arg_id);

void xfce_notify_log_gbus_emit_row_changed (
    XfceNotifyLogGBus *object,
    const gchar *arg_id);

void xfce_notify_log_gbus_emit_row_deleted (
    XfceNotifyLogGBus *object,
    const gchar *arg_id);

void xfce_notify_log_gbus_emit_truncated (
    XfceNotifyLogGBus *object,
    guint arg_n_entries_kept);

void xfce_notify_log_gbus_emit_cleared (
    XfceNotifyLogGBus *object);



/* D-Bus method calls: */
void xfce_notify_log_gbus_call_get (
    XfceNotifyLogGBus *proxy,
    const gchar *arg_id,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean xfce_notify_log_gbus_call_get_finish (
    XfceNotifyLogGBus *proxy,
    GVariant **out_entry,
    GAsyncResult *res,
    GError **error);

gboolean xfce_notify_log_gbus_call_get_sync (
    XfceNotifyLogGBus *proxy,
    const gchar *arg_id,
    GVariant **out_entry,
    GCancellable *cancellable,
    GError **error);

void xfce_notify_log_gbus_call_list (
    XfceNotifyLogGBus *proxy,
    const gchar *arg_start_after_id,
    guint arg_count,
    gboolean arg_only_unread,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean xfce_notify_log_gbus_call_list_finish (
    XfceNotifyLogGBus *proxy,
    GVariant **out_entries,
    GAsyncResult *res,
    GError **error);

gboolean xfce_notify_log_gbus_call_list_sync (
    XfceNotifyLogGBus *proxy,
    const gchar *arg_start_after_id,
    guint arg_count,
    gboolean arg_only_unread,
    GVariant **out_entries,
    GCancellable *cancellable,
    GError **error);

void xfce_notify_log_gbus_call_has_unread (
    XfceNotifyLogGBus *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean xfce_notify_log_gbus_call_has_unread_finish (
    XfceNotifyLogGBus *proxy,
    gboolean *out_has_unread,
    GAsyncResult *res,
    GError **error);

gboolean xfce_notify_log_gbus_call_has_unread_sync (
    XfceNotifyLogGBus *proxy,
    gboolean *out_has_unread,
    GCancellable *cancellable,
    GError **error);

void xfce_notify_log_gbus_call_get_app_id_counts (
    XfceNotifyLogGBus *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean xfce_notify_log_gbus_call_get_app_id_counts_finish (
    XfceNotifyLogGBus *proxy,
    GVariant **out_app_id_counts,
    GAsyncResult *res,
    GError **error);

gboolean xfce_notify_log_gbus_call_get_app_id_counts_sync (
    XfceNotifyLogGBus *proxy,
    GVariant **out_app_id_counts,
    GCancellable *cancellable,
    GError **error);

void xfce_notify_log_gbus_call_mark_read (
    XfceNotifyLogGBus *proxy,
    const gchar *const *arg_ids,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean xfce_notify_log_gbus_call_mark_read_finish (
    XfceNotifyLogGBus *proxy,
    GAsyncResult *res,
    GError **error);

gboolean xfce_notify_log_gbus_call_mark_read_sync (
    XfceNotifyLogGBus *proxy,
    const gchar *const *arg_ids,
    GCancellable *cancellable,
    GError **error);

void xfce_notify_log_gbus_call_mark_all_read (
    XfceNotifyLogGBus *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean xfce_notify_log_gbus_call_mark_all_read_finish (
    XfceNotifyLogGBus *proxy,
    GAsyncResult *res,
    GError **error);

gboolean xfce_notify_log_gbus_call_mark_all_read_sync (
    XfceNotifyLogGBus *proxy,
    GCancellable *cancellable,
    GError **error);

void xfce_notify_log_gbus_call_delete (
    XfceNotifyLogGBus *proxy,
    const gchar *const *arg_ids,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean xfce_notify_log_gbus_call_delete_finish (
    XfceNotifyLogGBus *proxy,
    GAsyncResult *res,
    GError **error);

gboolean xfce_notify_log_gbus_call_delete_sync (
    XfceNotifyLogGBus *proxy,
    const gchar *const *arg_ids,
    GCancellable *cancellable,
    GError **error);

void xfce_notify_log_gbus_call_truncate (
    XfceNotifyLogGBus *proxy,
    guint arg_entries_to_keep,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean xfce_notify_log_gbus_call_truncate_finish (
    XfceNotifyLogGBus *proxy,
    GAsyncResult *res,
    GError **error);

gboolean xfce_notify_log_gbus_call_truncate_sync (
    XfceNotifyLogGBus *proxy,
    guint arg_entries_to_keep,
    GCancellable *cancellable,
    GError **error);

void xfce_notify_log_gbus_call_clear (
    XfceNotifyLogGBus *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean xfce_notify_log_gbus_call_clear_finish (
    XfceNotifyLogGBus *proxy,
    GAsyncResult *res,
    GError **error);

gboolean xfce_notify_log_gbus_call_clear_sync (
    XfceNotifyLogGBus *proxy,
    GCancellable *cancellable,
    GError **error);



/* ---- */

#define XFCE_TYPE_NOTIFY_LOG_GBUS_PROXY (xfce_notify_log_gbus_proxy_get_type ())
#define XFCE_NOTIFY_LOG_GBUS_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XFCE_TYPE_NOTIFY_LOG_GBUS_PROXY, XfceNotifyLogGBusProxy))
#define XFCE_NOTIFY_LOG_GBUS_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), XFCE_TYPE_NOTIFY_LOG_GBUS_PROXY, XfceNotifyLogGBusProxyClass))
#define XFCE_NOTIFY_LOG_GBUS_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), XFCE_TYPE_NOTIFY_LOG_GBUS_PROXY, XfceNotifyLogGBusProxyClass))
#define XFCE_IS_NOTIFY_LOG_GBUS_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XFCE_TYPE_NOTIFY_LOG_GBUS_PROXY))
#define XFCE_IS_NOTIFY_LOG_GBUS_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), XFCE_TYPE_NOTIFY_LOG_GBUS_PROXY))

typedef struct _XfceNotifyLogGBusProxy XfceNotifyLogGBusProxy;
typedef struct _XfceNotifyLogGBusProxyClass XfceNotifyLogGBusProxyClass;
typedef struct _XfceNotifyLogGBusProxyPrivate XfceNotifyLogGBusProxyPrivate;

struct _XfceNotifyLogGBusProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  XfceNotifyLogGBusProxyPrivate *priv;
};

struct _XfceNotifyLogGBusProxyClass
{
  GDBusProxyClass parent_class;
};

GType xfce_notify_log_gbus_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (XfceNotifyLogGBusProxy, g_object_unref)
#endif

void xfce_notify_log_gbus_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
XfceNotifyLogGBus *xfce_notify_log_gbus_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
XfceNotifyLogGBus *xfce_notify_log_gbus_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void xfce_notify_log_gbus_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
XfceNotifyLogGBus *xfce_notify_log_gbus_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
XfceNotifyLogGBus *xfce_notify_log_gbus_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define XFCE_TYPE_NOTIFY_LOG_GBUS_SKELETON (xfce_notify_log_gbus_skeleton_get_type ())
#define XFCE_NOTIFY_LOG_GBUS_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XFCE_TYPE_NOTIFY_LOG_GBUS_SKELETON, XfceNotifyLogGBusSkeleton))
#define XFCE_NOTIFY_LOG_GBUS_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), XFCE_TYPE_NOTIFY_LOG_GBUS_SKELETON, XfceNotifyLogGBusSkeletonClass))
#define XFCE_NOTIFY_LOG_GBUS_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), XFCE_TYPE_NOTIFY_LOG_GBUS_SKELETON, XfceNotifyLogGBusSkeletonClass))
#define XFCE_IS_NOTIFY_LOG_GBUS_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XFCE_TYPE_NOTIFY_LOG_GBUS_SKELETON))
#define XFCE_IS_NOTIFY_LOG_GBUS_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), XFCE_TYPE_NOTIFY_LOG_GBUS_SKELETON))

typedef struct _XfceNotifyLogGBusSkeleton XfceNotifyLogGBusSkeleton;
typedef struct _XfceNotifyLogGBusSkeletonClass XfceNotifyLogGBusSkeletonClass;
typedef struct _XfceNotifyLogGBusSkeletonPrivate XfceNotifyLogGBusSkeletonPrivate;

struct _XfceNotifyLogGBusSkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  XfceNotifyLogGBusSkeletonPrivate *priv;
};

struct _XfceNotifyLogGBusSkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType xfce_notify_log_gbus_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (XfceNotifyLogGBusSkeleton, g_object_unref)
#endif

XfceNotifyLogGBus *xfce_notify_log_gbus_skeleton_new (void);


G_END_DECLS

#endif /* __XFCE_NOTIFY_LOG_GBUS_H__ */
