#XfceNotifyWindow {
    background-color: #ff69b4;
    border-radius: 10px;
    border: 6px solid #ffff00;
}

#XfceNotifyWindow label,
#XfceNotifyWindow image {
    font-family: Comic Sans MS;
    font-size: 22px;
    color: #ffff00;
}

#XfceNotifyWindow label#summary {
    font-weight: Bold;
}

#XfceNotifyWindow button {
    border-style: none;
    background-image: none;
    background-color : #9400d3;
}

#XfceNotifyWindow button:hover {
    border-style: none;
    background-image: none;
    background-color : #5e0086;
}

#XfceNotifyWindow progressbar {
    border-radius: 3px;
}

#XfceNotifyWindow progressbar progress {
    min-height: 16px;
    background-image: none;
    background-color: shade(#ffff00, 1.2);
}

#XfceNotifyWindow progressbar trough {
    min-height: 16px;
    background-image: none;
    background-color: #5e0086;
    border: 1px solid shade(#5e0086, 0.6);
    border-radius: 3px;
}
