/*
 * This file is generated by gdbus-codegen, do not modify it.
 *
 * The license of this code is the same as for the D-Bus interface description
 * it was derived from. Note that it links to GLib, so must comply with the
 * LGPL linking clauses.
 */

#ifndef __XFCE_NOTIFY_FDO_GBUS_H__
#define __XFCE_NOTIFY_FDO_GBUS_H__

#include <gio/gio.h>

G_BEGIN_DECLS


/* ------------------------------------------------------------------------ */
/* Declarations for org.freedesktop.Notifications */

#define XFCE_TYPE_NOTIFY_FDO_GBUS (xfce_notify_fdo_gbus_get_type ())
#define XFCE_NOTIFY_FDO_GBUS(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XFCE_TYPE_NOTIFY_FDO_GBUS, XfceNotifyFdoGBus))
#define XFCE_IS_NOTIFY_FDO_GBUS(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XFCE_TYPE_NOTIFY_FDO_GBUS))
#define XFCE_NOTIFY_FDO_GBUS_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), XFCE_TYPE_NOTIFY_FDO_GBUS, XfceNotifyFdoGBusIface))

struct _XfceNotifyFdoGBus;
typedef struct _XfceNotifyFdoGBus XfceNotifyFdoGBus;
typedef struct _XfceNotifyFdoGBusIface XfceNotifyFdoGBusIface;

struct _XfceNotifyFdoGBusIface
{
  GTypeInterface parent_iface;


  gboolean (*handle_close_notification) (
    XfceNotifyFdoGBus *object,
    GDBusMethodInvocation *invocation,
    guint arg_id);

  gboolean (*handle_get_capabilities) (
    XfceNotifyFdoGBus *object,
    GDBusMethodInvocation *invocation);

  gboolean (*handle_get_server_information) (
    XfceNotifyFdoGBus *object,
    GDBusMethodInvocation *invocation);

  gboolean (*handle_notify) (
    XfceNotifyFdoGBus *object,
    GDBusMethodInvocation *invocation,
    const gchar *arg_app_name,
    guint arg_replaces_id,
    const gchar *arg_app_icon,
    const gchar *arg_summary,
    const gchar *arg_body,
    const gchar *const *arg_actions,
    GVariant *arg_hints,
    gint arg_expire_timeout);

  void (*action_invoked) (
    XfceNotifyFdoGBus *object,
    guint arg_id,
    const gchar *arg_action_key);

  void (*notification_closed) (
    XfceNotifyFdoGBus *object,
    guint arg_id,
    guint arg_reason);

};

GType xfce_notify_fdo_gbus_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *xfce_notify_fdo_gbus_interface_info (void);
guint xfce_notify_fdo_gbus_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
void xfce_notify_fdo_gbus_complete_get_capabilities (
    XfceNotifyFdoGBus *object,
    GDBusMethodInvocation *invocation,
    const gchar *const *capabilities);

void xfce_notify_fdo_gbus_complete_notify (
    XfceNotifyFdoGBus *object,
    GDBusMethodInvocation *invocation,
    guint id);

void xfce_notify_fdo_gbus_complete_close_notification (
    XfceNotifyFdoGBus *object,
    GDBusMethodInvocation *invocation);

void xfce_notify_fdo_gbus_complete_get_server_information (
    XfceNotifyFdoGBus *object,
    GDBusMethodInvocation *invocation,
    const gchar *name,
    const gchar *vendor,
    const gchar *version,
    const gchar *spec_version);



/* D-Bus signal emissions functions: */
void xfce_notify_fdo_gbus_emit_notification_closed (
    XfceNotifyFdoGBus *object,
    guint arg_id,
    guint arg_reason);

void xfce_notify_fdo_gbus_emit_action_invoked (
    XfceNotifyFdoGBus *object,
    guint arg_id,
    const gchar *arg_action_key);



/* D-Bus method calls: */
void xfce_notify_fdo_gbus_call_get_capabilities (
    XfceNotifyFdoGBus *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean xfce_notify_fdo_gbus_call_get_capabilities_finish (
    XfceNotifyFdoGBus *proxy,
    gchar ***out_capabilities,
    GAsyncResult *res,
    GError **error);

gboolean xfce_notify_fdo_gbus_call_get_capabilities_sync (
    XfceNotifyFdoGBus *proxy,
    gchar ***out_capabilities,
    GCancellable *cancellable,
    GError **error);

void xfce_notify_fdo_gbus_call_notify (
    XfceNotifyFdoGBus *proxy,
    const gchar *arg_app_name,
    guint arg_replaces_id,
    const gchar *arg_app_icon,
    const gchar *arg_summary,
    const gchar *arg_body,
    const gchar *const *arg_actions,
    GVariant *arg_hints,
    gint arg_expire_timeout,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean xfce_notify_fdo_gbus_call_notify_finish (
    XfceNotifyFdoGBus *proxy,
    guint *out_id,
    GAsyncResult *res,
    GError **error);

gboolean xfce_notify_fdo_gbus_call_notify_sync (
    XfceNotifyFdoGBus *proxy,
    const gchar *arg_app_name,
    guint arg_replaces_id,
    const gchar *arg_app_icon,
    const gchar *arg_summary,
    const gchar *arg_body,
    const gchar *const *arg_actions,
    GVariant *arg_hints,
    gint arg_expire_timeout,
    guint *out_id,
    GCancellable *cancellable,
    GError **error);

void xfce_notify_fdo_gbus_call_close_notification (
    XfceNotifyFdoGBus *proxy,
    guint arg_id,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean xfce_notify_fdo_gbus_call_close_notification_finish (
    XfceNotifyFdoGBus *proxy,
    GAsyncResult *res,
    GError **error);

gboolean xfce_notify_fdo_gbus_call_close_notification_sync (
    XfceNotifyFdoGBus *proxy,
    guint arg_id,
    GCancellable *cancellable,
    GError **error);

void xfce_notify_fdo_gbus_call_get_server_information (
    XfceNotifyFdoGBus *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

gboolean xfce_notify_fdo_gbus_call_get_server_information_finish (
    XfceNotifyFdoGBus *proxy,
    gchar **out_name,
    gchar **out_vendor,
    gchar **out_version,
    gchar **out_spec_version,
    GAsyncResult *res,
    GError **error);

gboolean xfce_notify_fdo_gbus_call_get_server_information_sync (
    XfceNotifyFdoGBus *proxy,
    gchar **out_name,
    gchar **out_vendor,
    gchar **out_version,
    gchar **out_spec_version,
    GCancellable *cancellable,
    GError **error);



/* ---- */

#define XFCE_TYPE_NOTIFY_FDO_GBUS_PROXY (xfce_notify_fdo_gbus_proxy_get_type ())
#define XFCE_NOTIFY_FDO_GBUS_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XFCE_TYPE_NOTIFY_FDO_GBUS_PROXY, XfceNotifyFdoGBusProxy))
#define XFCE_NOTIFY_FDO_GBUS_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), XFCE_TYPE_NOTIFY_FDO_GBUS_PROXY, XfceNotifyFdoGBusProxyClass))
#define XFCE_NOTIFY_FDO_GBUS_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), XFCE_TYPE_NOTIFY_FDO_GBUS_PROXY, XfceNotifyFdoGBusProxyClass))
#define XFCE_IS_NOTIFY_FDO_GBUS_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XFCE_TYPE_NOTIFY_FDO_GBUS_PROXY))
#define XFCE_IS_NOTIFY_FDO_GBUS_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), XFCE_TYPE_NOTIFY_FDO_GBUS_PROXY))

typedef struct _XfceNotifyFdoGBusProxy XfceNotifyFdoGBusProxy;
typedef struct _XfceNotifyFdoGBusProxyClass XfceNotifyFdoGBusProxyClass;
typedef struct _XfceNotifyFdoGBusProxyPrivate XfceNotifyFdoGBusProxyPrivate;

struct _XfceNotifyFdoGBusProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  XfceNotifyFdoGBusProxyPrivate *priv;
};

struct _XfceNotifyFdoGBusProxyClass
{
  GDBusProxyClass parent_class;
};

GType xfce_notify_fdo_gbus_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (XfceNotifyFdoGBusProxy, g_object_unref)
#endif

void xfce_notify_fdo_gbus_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
XfceNotifyFdoGBus *xfce_notify_fdo_gbus_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
XfceNotifyFdoGBus *xfce_notify_fdo_gbus_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void xfce_notify_fdo_gbus_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
XfceNotifyFdoGBus *xfce_notify_fdo_gbus_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
XfceNotifyFdoGBus *xfce_notify_fdo_gbus_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define XFCE_TYPE_NOTIFY_FDO_GBUS_SKELETON (xfce_notify_fdo_gbus_skeleton_get_type ())
#define XFCE_NOTIFY_FDO_GBUS_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XFCE_TYPE_NOTIFY_FDO_GBUS_SKELETON, XfceNotifyFdoGBusSkeleton))
#define XFCE_NOTIFY_FDO_GBUS_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), XFCE_TYPE_NOTIFY_FDO_GBUS_SKELETON, XfceNotifyFdoGBusSkeletonClass))
#define XFCE_NOTIFY_FDO_GBUS_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), XFCE_TYPE_NOTIFY_FDO_GBUS_SKELETON, XfceNotifyFdoGBusSkeletonClass))
#define XFCE_IS_NOTIFY_FDO_GBUS_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XFCE_TYPE_NOTIFY_FDO_GBUS_SKELETON))
#define XFCE_IS_NOTIFY_FDO_GBUS_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), XFCE_TYPE_NOTIFY_FDO_GBUS_SKELETON))

typedef struct _XfceNotifyFdoGBusSkeleton XfceNotifyFdoGBusSkeleton;
typedef struct _XfceNotifyFdoGBusSkeletonClass XfceNotifyFdoGBusSkeletonClass;
typedef struct _XfceNotifyFdoGBusSkeletonPrivate XfceNotifyFdoGBusSkeletonPrivate;

struct _XfceNotifyFdoGBusSkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  XfceNotifyFdoGBusSkeletonPrivate *priv;
};

struct _XfceNotifyFdoGBusSkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType xfce_notify_fdo_gbus_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (XfceNotifyFdoGBusSkeleton, g_object_unref)
#endif

XfceNotifyFdoGBus *xfce_notify_fdo_gbus_skeleton_new (void);


/* ------------------------------------------------------------------------ */
/* Declarations for org.xfce.Notifyd */

#define XFCE_TYPE_NOTIFY_OLD_GBUS (xfce_notify_old_gbus_get_type ())
#define XFCE_NOTIFY_OLD_GBUS(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XFCE_TYPE_NOTIFY_OLD_GBUS, XfceNotifyOldGBus))
#define XFCE_IS_NOTIFY_OLD_GBUS(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XFCE_TYPE_NOTIFY_OLD_GBUS))
#define XFCE_NOTIFY_OLD_GBUS_GET_IFACE(o) (G_TYPE_INSTANCE_GET_INTERFACE ((o), XFCE_TYPE_NOTIFY_OLD_GBUS, XfceNotifyOldGBusIface))

struct _XfceNotifyOldGBus;
typedef struct _XfceNotifyOldGBus XfceNotifyOldGBus;
typedef struct _XfceNotifyOldGBusIface XfceNotifyOldGBusIface;

struct _XfceNotifyOldGBusIface
{
  GTypeInterface parent_iface;

  gboolean (*handle_quit) (
    XfceNotifyOldGBus *object,
    GDBusMethodInvocation *invocation);

};

GType xfce_notify_old_gbus_get_type (void) G_GNUC_CONST;

GDBusInterfaceInfo *xfce_notify_old_gbus_interface_info (void);
guint xfce_notify_old_gbus_override_properties (GObjectClass *klass, guint property_id_begin);


/* D-Bus method call completion functions: */
G_GNUC_DEPRECATED void xfce_notify_old_gbus_complete_quit (
    XfceNotifyOldGBus *object,
    GDBusMethodInvocation *invocation);



/* D-Bus method calls: */
G_GNUC_DEPRECATED void xfce_notify_old_gbus_call_quit (
    XfceNotifyOldGBus *proxy,
    GCancellable *cancellable,
    GAsyncReadyCallback callback,
    gpointer user_data);

G_GNUC_DEPRECATED gboolean xfce_notify_old_gbus_call_quit_finish (
    XfceNotifyOldGBus *proxy,
    GAsyncResult *res,
    GError **error);

G_GNUC_DEPRECATED gboolean xfce_notify_old_gbus_call_quit_sync (
    XfceNotifyOldGBus *proxy,
    GCancellable *cancellable,
    GError **error);



/* ---- */

#define XFCE_TYPE_NOTIFY_OLD_GBUS_PROXY (xfce_notify_old_gbus_proxy_get_type ())
#define XFCE_NOTIFY_OLD_GBUS_PROXY(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XFCE_TYPE_NOTIFY_OLD_GBUS_PROXY, XfceNotifyOldGBusProxy))
#define XFCE_NOTIFY_OLD_GBUS_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), XFCE_TYPE_NOTIFY_OLD_GBUS_PROXY, XfceNotifyOldGBusProxyClass))
#define XFCE_NOTIFY_OLD_GBUS_PROXY_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), XFCE_TYPE_NOTIFY_OLD_GBUS_PROXY, XfceNotifyOldGBusProxyClass))
#define XFCE_IS_NOTIFY_OLD_GBUS_PROXY(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XFCE_TYPE_NOTIFY_OLD_GBUS_PROXY))
#define XFCE_IS_NOTIFY_OLD_GBUS_PROXY_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), XFCE_TYPE_NOTIFY_OLD_GBUS_PROXY))

typedef struct _XfceNotifyOldGBusProxy XfceNotifyOldGBusProxy;
typedef struct _XfceNotifyOldGBusProxyClass XfceNotifyOldGBusProxyClass;
typedef struct _XfceNotifyOldGBusProxyPrivate XfceNotifyOldGBusProxyPrivate;

struct _XfceNotifyOldGBusProxy
{
  /*< private >*/
  GDBusProxy parent_instance;
  XfceNotifyOldGBusProxyPrivate *priv;
};

struct _XfceNotifyOldGBusProxyClass
{
  GDBusProxyClass parent_class;
};

GType xfce_notify_old_gbus_proxy_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (XfceNotifyOldGBusProxy, g_object_unref)
#endif

void xfce_notify_old_gbus_proxy_new (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
XfceNotifyOldGBus *xfce_notify_old_gbus_proxy_new_finish (
    GAsyncResult        *res,
    GError             **error);
XfceNotifyOldGBus *xfce_notify_old_gbus_proxy_new_sync (
    GDBusConnection     *connection,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);

void xfce_notify_old_gbus_proxy_new_for_bus (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GAsyncReadyCallback  callback,
    gpointer             user_data);
XfceNotifyOldGBus *xfce_notify_old_gbus_proxy_new_for_bus_finish (
    GAsyncResult        *res,
    GError             **error);
XfceNotifyOldGBus *xfce_notify_old_gbus_proxy_new_for_bus_sync (
    GBusType             bus_type,
    GDBusProxyFlags      flags,
    const gchar         *name,
    const gchar         *object_path,
    GCancellable        *cancellable,
    GError             **error);


/* ---- */

#define XFCE_TYPE_NOTIFY_OLD_GBUS_SKELETON (xfce_notify_old_gbus_skeleton_get_type ())
#define XFCE_NOTIFY_OLD_GBUS_SKELETON(o) (G_TYPE_CHECK_INSTANCE_CAST ((o), XFCE_TYPE_NOTIFY_OLD_GBUS_SKELETON, XfceNotifyOldGBusSkeleton))
#define XFCE_NOTIFY_OLD_GBUS_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_CAST ((k), XFCE_TYPE_NOTIFY_OLD_GBUS_SKELETON, XfceNotifyOldGBusSkeletonClass))
#define XFCE_NOTIFY_OLD_GBUS_SKELETON_GET_CLASS(o) (G_TYPE_INSTANCE_GET_CLASS ((o), XFCE_TYPE_NOTIFY_OLD_GBUS_SKELETON, XfceNotifyOldGBusSkeletonClass))
#define XFCE_IS_NOTIFY_OLD_GBUS_SKELETON(o) (G_TYPE_CHECK_INSTANCE_TYPE ((o), XFCE_TYPE_NOTIFY_OLD_GBUS_SKELETON))
#define XFCE_IS_NOTIFY_OLD_GBUS_SKELETON_CLASS(k) (G_TYPE_CHECK_CLASS_TYPE ((k), XFCE_TYPE_NOTIFY_OLD_GBUS_SKELETON))

typedef struct _XfceNotifyOldGBusSkeleton XfceNotifyOldGBusSkeleton;
typedef struct _XfceNotifyOldGBusSkeletonClass XfceNotifyOldGBusSkeletonClass;
typedef struct _XfceNotifyOldGBusSkeletonPrivate XfceNotifyOldGBusSkeletonPrivate;

struct _XfceNotifyOldGBusSkeleton
{
  /*< private >*/
  GDBusInterfaceSkeleton parent_instance;
  XfceNotifyOldGBusSkeletonPrivate *priv;
};

struct _XfceNotifyOldGBusSkeletonClass
{
  GDBusInterfaceSkeletonClass parent_class;
};

GType xfce_notify_old_gbus_skeleton_get_type (void) G_GNUC_CONST;

#if GLIB_CHECK_VERSION(2, 44, 0)
G_DEFINE_AUTOPTR_CLEANUP_FUNC (XfceNotifyOldGBusSkeleton, g_object_unref)
#endif

XfceNotifyOldGBus *xfce_notify_old_gbus_skeleton_new (void);


G_END_DECLS

#endif /* __XFCE_NOTIFY_FDO_GBUS_H__ */
