
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#ifndef __XFCE_NOTIFY_ENUM_TYPES_H__
#define __XFCE_NOTIFY_ENUM_TYPES_H__

#include <glib-object.h>

G_BEGIN_DECLS
/* enumerations from "./xfce-notify-common.h" */
GType xfce_notify_display_fields_get_type(void) G_GNUC_CONST;
#define XFCE_TYPE_NOTIFY_DISPLAY_FIELDS (xfce_notify_display_fields_get_type())
GType xfce_notify_show_on_get_type(void) G_GNUC_CONST;
#define XFCE_TYPE_NOTIFY_SHOW_ON (xfce_notify_show_on_get_type())
GType xfce_notify_position_get_type(void) G_GNUC_CONST;
#define XFCE_TYPE_NOTIFY_POSITION (xfce_notify_position_get_type())
GType xfce_notify_datetime_format_get_type(void) G_GNUC_CONST;
#define XFCE_TYPE_NOTIFY_DATETIME_FORMAT (xfce_notify_datetime_format_get_type())
GType xfce_log_level_get_type(void) G_GNUC_CONST;
#define XFCE_TYPE_LOG_LEVEL (xfce_log_level_get_type())
GType xfce_log_level_apps_get_type(void) G_GNUC_CONST;
#define XFCE_TYPE_LOG_LEVEL_APPS (xfce_log_level_apps_get_type())
GType xfce_notify_urgency_get_type(void) G_GNUC_CONST;
#define XFCE_TYPE_NOTIFY_URGENCY (xfce_notify_urgency_get_type())
GType xfce_notify_close_reason_get_type(void) G_GNUC_CONST;
#define XFCE_TYPE_NOTIFY_CLOSE_REASON (xfce_notify_close_reason_get_type())
G_END_DECLS

#endif /* !__XFCE_NOTIFY_ENUM_TYPES_H__ */

/* Generated data ends here */

