
/* This file is generated by glib-mkenums, do not modify it. This code is licensed under the same license as the containing project. Note that it links to GLib, so must comply with the LGPL linking clauses. */

#include "xfce-notify-common.h"
#include "xfce-notify-enum-types.h"

/* enumerations from "./xfce-notify-common.h" */
GType
xfce_notify_display_fields_get_type(void)
{
    static GType type = 0;
    if (type == 0) {
        static const GEnumValue values[] = {
            { XFCE_NOTIFY_DISPLAY_FULL, "XFCE_NOTIFY_DISPLAY_FULL", "full" },
            { XFCE_NOTIFY_DISPLAY_ICON_SUMMARY, "XFCE_NOTIFY_DISPLAY_ICON_SUMMARY", "icon-summary" },
            { XFCE_NOTIFY_DISPLAY_ICON_APPNAME, "XFCE_NOTIFY_DISPLAY_ICON_APPNAME", "icon-appname" },
            { 0, NULL, NULL }
        };
       type = g_enum_register_static("XfceNotifyDisplayFields", values);
    }
    return type;
}
GType
xfce_notify_show_on_get_type(void)
{
    static GType type = 0;
    if (type == 0) {
        static const GEnumValue values[] = {
            { XFCE_NOTIFY_SHOW_ON_ACTIVE_MONITOR, "XFCE_NOTIFY_SHOW_ON_ACTIVE_MONITOR", "active-monitor" },
            { XFCE_NOTIFY_SHOW_ON_PRIMARY_MONITOR, "XFCE_NOTIFY_SHOW_ON_PRIMARY_MONITOR", "primary-monitor" },
            { XFCE_NOTIFY_SHOW_ON_ALL_MONITORS, "XFCE_NOTIFY_SHOW_ON_ALL_MONITORS", "all-monitors" },
            { 0, NULL, NULL }
        };
       type = g_enum_register_static("XfceNotifyShowOn", values);
    }
    return type;
}
GType
xfce_notify_position_get_type(void)
{
    static GType type = 0;
    if (type == 0) {
        static const GEnumValue values[] = {
            { XFCE_NOTIFY_POS_TOP_LEFT, "XFCE_NOTIFY_POS_TOP_LEFT", "top-left" },
            { XFCE_NOTIFY_POS_BOTTOM_LEFT, "XFCE_NOTIFY_POS_BOTTOM_LEFT", "bottom-left" },
            { XFCE_NOTIFY_POS_TOP_RIGHT, "XFCE_NOTIFY_POS_TOP_RIGHT", "top-right" },
            { XFCE_NOTIFY_POS_BOTTOM_RIGHT, "XFCE_NOTIFY_POS_BOTTOM_RIGHT", "bottom-right" },
            { XFCE_NOTIFY_POS_TOP_CENTER, "XFCE_NOTIFY_POS_TOP_CENTER", "top-center" },
            { XFCE_NOTIFY_POS_BOTTOM_CENTER, "XFCE_NOTIFY_POS_BOTTOM_CENTER", "bottom-center" },
            { 0, NULL, NULL }
        };
       type = g_enum_register_static("XfceNotifyPosition", values);
    }
    return type;
}
GType
xfce_notify_datetime_format_get_type(void)
{
    static GType type = 0;
    if (type == 0) {
        static const GEnumValue values[] = {
            { XFCE_NOTIFY_DATETIME_LOCALE_DEFAULT, "XFCE_NOTIFY_DATETIME_LOCALE_DEFAULT", "locale-default" },
            { XFCE_NOTIFY_DATETIME_RELATIVE_TIMES, "XFCE_NOTIFY_DATETIME_RELATIVE_TIMES", "relative-times" },
            { XFCE_NOTIFY_DATETIME_ISO8601, "XFCE_NOTIFY_DATETIME_ISO8601", "iso8601" },
            { XFCE_NOTIFY_DATETIME_CUSTOM, "XFCE_NOTIFY_DATETIME_CUSTOM", "custom" },
            { 0, NULL, NULL }
        };
       type = g_enum_register_static("XfceNotifyDatetimeFormat", values);
    }
    return type;
}
GType
xfce_log_level_get_type(void)
{
    static GType type = 0;
    if (type == 0) {
        static const GEnumValue values[] = {
            { XFCE_LOG_LEVEL_NOT_FULLY_SHOWN, "XFCE_LOG_LEVEL_NOT_FULLY_SHOWN", "not-fully-shown" },
            { XFCE_LOG_LEVEL_ALWAYS, "XFCE_LOG_LEVEL_ALWAYS", "always" },
            { 0, NULL, NULL }
        };
       type = g_enum_register_static("XfceLogLevel", values);
    }
    return type;
}
GType
xfce_log_level_apps_get_type(void)
{
    static GType type = 0;
    if (type == 0) {
        static const GEnumValue values[] = {
            { XFCE_LOG_LEVEL_APPS_ALL, "XFCE_LOG_LEVEL_APPS_ALL", "all" },
            { XFCE_LOG_LEVEL_APPS_EXCEPT_BLOCKED, "XFCE_LOG_LEVEL_APPS_EXCEPT_BLOCKED", "except-blocked" },
            { XFCE_LOG_LEVEL_APPS_ONLY_BLOCKED, "XFCE_LOG_LEVEL_APPS_ONLY_BLOCKED", "only-blocked" },
            { 0, NULL, NULL }
        };
       type = g_enum_register_static("XfceLogLevelApps", values);
    }
    return type;
}
GType
xfce_notify_urgency_get_type(void)
{
    static GType type = 0;
    if (type == 0) {
        static const GEnumValue values[] = {
            { XFCE_NOTIFY_URGENCY_LOW, "XFCE_NOTIFY_URGENCY_LOW", "low" },
            { XFCE_NOTIFY_URGENCY_NORMAL, "XFCE_NOTIFY_URGENCY_NORMAL", "normal" },
            { XFCE_NOTIFY_URGENCY_CRITICAL, "XFCE_NOTIFY_URGENCY_CRITICAL", "critical" },
            { 0, NULL, NULL }
        };
       type = g_enum_register_static("XfceNotifyUrgency", values);
    }
    return type;
}
GType
xfce_notify_close_reason_get_type(void)
{
    static GType type = 0;
    if (type == 0) {
        static const GEnumValue values[] = {
            { XFCE_NOTIFY_CLOSE_REASON_EXPIRED, "XFCE_NOTIFY_CLOSE_REASON_EXPIRED", "expired" },
            { XFCE_NOTIFY_CLOSE_REASON_DISMISSED, "XFCE_NOTIFY_CLOSE_REASON_DISMISSED", "dismissed" },
            { XFCE_NOTIFY_CLOSE_REASON_CLIENT, "XFCE_NOTIFY_CLOSE_REASON_CLIENT", "client" },
            { XFCE_NOTIFY_CLOSE_REASON_UNKNOWN, "XFCE_NOTIFY_CLOSE_REASON_UNKNOWN", "unknown" },
            { 0, NULL, NULL }
        };
       type = g_enum_register_static("XfceNotifyCloseReason", values);
    }
    return type;
}

/* Generated data ends here */

