/* We set some default options here.  This is loaded with a low priority,
 * the rest will use whatever default colors are in your main gtk theme */
#XfceNotifyWindow {
    border-radius: 6px;
    border-width: 1px;
    border-color : #ffffff;
}

#XfceNotifyWindow label#summary {
    font-weight: Bold;
}
