#XfceNotifyWindow {
    background-color: #f2f1f1;
    border-radius: 10px;
    border: 1px solid #686868;
    padding: 20px;
}

#XfceNotifyWindow:hover {
    background-color: shade(#f2f1f1, 0.98);
}

#XfceNotifyWindow label,
#XfceNotifyWindow image {
    color: #686868;
}

#XfceNotifyWindow label#summary {
    font-weight: Bold;
}

#XfceNotifyWindow button {
    font-weight: Bold;
    border: 1px solid shade(#686868, 1.15);
    box-shadow: none;
    background-image: none;
    background-color: #f2f1f1;
    color: #686868;
}

#XfceNotifyWindow button:hover {
    box-shadow: none;
    background-image: none;
    background-color: #e4e4e4;
    color: #686868;
}

#XfceNotifyWindow progressbar {
    min-height: 10px;
    border-radius: 3px;
}

#XfceNotifyWindow progressbar progress {
    background-image: none;
    background-color: shade (#ddd, 0.65);
    border: 1px solid shade (#ddd, 0.55);
}

#XfceNotifyWindow progressbar trough {
    background-image: linear-gradient(to bottom,
                                      shade(#ddd, 0.98),
                                      shade(#ddd, 1.25)
                                      );
    border: 1px solid shade(#ddd, 0.8);
    border-radius: 3px;
}
